from config import PNL_GENERAL_THRESHOLD
from enum import Enum, auto
import json

class AutoName(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

class Intent(AutoName):
    CREATE_OR_UPDATE = auto()
    DELETE = auto()
    GREET = auto()
    READ = auto()
    SAY_GOODBYE = auto()

class Entity(AutoName):#TODO: #15 to change name to differ from entity.py
    ATTRIBUTE = auto()
    CLASS = auto()
    CONTACT = auto()
    EMAIL = auto()
    LOCATION = auto()

    def update(self, body, role) -> None:
        self.body = body
        self.role = role
    
class Attribute():#TODO: #14 to change name to differ from attribute.py
    def __init__(self, ) -> None:
        pass

class ParseResponse:
    def __init__(self, response) -> None:
        self.__response = response
        self.__intent = None
        self.__entities = []
        
        if (len(response['intents'])>0
            and response['intents'][0]['confidence'] > PNL_GENERAL_THRESHOLD):
                self.__intent = Intent(response['intents'][0]['name'].upper())

        #print(json.dumps(response, indent=3))
        
        for key in response['entities']:
            for entity in response['entities'][key]:
                if entity['confidence'] > PNL_GENERAL_THRESHOLD:
                    new_ent = Entity(entity['name'].replace('wit$','').upper())
                    new_ent.update(entity['body'], entity['role'])
                    self.__entities.append(new_ent)

    def getIntent(self) -> Intent:
        return self.__intent

    def intentIs(self, intent) -> bool:
        return self.getIntent() == intent
        
    def intentIs_GREET(self) -> bool:
        return self.intentIs(Intent.GREET)
    
    def intentIs_CREATE_OR_UPDATE(self) -> bool:
        return self.intentIs(Intent.CREATE_OR_UPDATE)

    def intentIs_DELETE(self) -> bool:
        return self.intentIs(Intent.DELETE)

    def intentIs_READ(self) -> bool:
        return self.intentIs(Intent.READ)

    def intentIs_SAY_GOODBYE(self) -> bool:
        return self.intentIs(Intent.SAY_GOODBYE)

    def getEntities(self) -> list:
        return self.__entities
    
    def getEntitiesByType(self, entityType):
        listReturn = []
        for entity in self.getEntities():
            if entity == entityType:
                listReturn.append(entity)
        return listReturn
    
    def getEntities_CLASS(self):
        return self.getEntitiesByType(Entity.CLASS)
    
    def getEntities_ATTRIBUTE(self):
        return self.getEntitiesByType(Entity.ATTRIBUTE)    


from src.text2system.auxiliary.constants import OPR_ATTRIBUTE_ADD, OPR_ENTITY_ADD
from src.text2system.autonomouscontroller import AutonomousController
from src.text2system.integrationengine import IntegrationEngine
from config import *

class SecurityEngine:
    def __init__(self, MUP, IE=None):
        self.__MUP = MUP #Multchannel App
        
        if IE is None:
            self.__IE = IntegrationEngine(self)
        else:
            self.__IE = IE #Integration Engine
        
        #TODO: security operations
        #...
        #user access allowed        
        self.__AC = AutonomousController(self) #autonomous controller instance
    def __authorize(self, opr):
        return True #for this experiment, all operations will be allowed

    def execute(self, opr, data={}):
        if not(self.__authorize(opr)):
            return {'error': 'authorization error'}
        #else: authorized
        #call Autonomous Controller
        return self.__AC.plan(opr, data)

    #util methods
    def getUser(self):
        return self.__MUP.user



from src.text2system.auxiliary.entity import Entity

class DomainEngine:
    def __init__(self, AC):
        self.__AC = AC #Autonomous Controller Object
        self.__entities = []

    def addEntity(self, name):
        #TODO: update meta data (MDB) and Transaction Data (TDB)
        e = Entity(name)
        if self.__entities.count(name) == 0:
            self.__entities.append(e)
        return e

    def getEntities(self):
        return self.__entities

    def addAttribute(self, entity, name, type, notnull=False):
        #TODO: #2 update meta data (MDB) and Transaction Data (TDB)
        #...
        entity.addAttribute(name, type, notnull)
        return True




import random
from src.text2system.auxiliary.constants import OPR_APP_HOME_CMD, OPR_APP_HOME_WEB, OPR_ATTRIBUTE_ADD, OPR_BOT_HANDLE_GET, OPR_ENTITY_ADD
from src.text2system.aiengine import AIEngine
from src.text2system.interfacecontroller import InterfaceController
from src.text2system.domainengine import DomainEngine
from config import *
from src.text2system.auxiliary.responseParser import *

class AutonomousController:
    def __init__(self, SE):
        self.__SE = SE #Security Engine object
        self.__DE = DomainEngine(self) #Domain Engine object
        self.__IC = InterfaceController(self) #Interface Controller object
        self.__AIE = AIEngine() #Artificial Intelligence Engine object
        self.__lastChatDth = None

    def __monitor(self):
        pass

    def __analyze(self):
        pass
    
    def plan(self, opr, data):
        #in this version, all tasks are going to be executed immediately
        return self.__execute(opr, data) 
    
    def __execute(self, opr, data):
        #TODO: manager the type of task
        #...
        if opr == OPR_APP_HOME_WEB:
            self.__IC.updateAppWeb()
            return {'homeurl': WEBAPP_HOME_URL}
        elif opr == OPR_APP_HOME_CMD:
            self.__IC.getApp_cmd(self.app_cmd_msgHandle)
            return True #TODO: to analyse return type/value
        elif opr == OPR_ENTITY_ADD:
            return self.__DE.addEntity(data['name'])
            #return True #TODO: #3 analysing return type
        elif opr == OPR_ATTRIBUTE_ADD:
            self.__DE.addAttribute(data['entity'], data['name']
                                   , data['type'], data['notnull'])
            self.__IC.updateAppWeb()
            return True
        elif opr == OPR_BOT_HANDLE_GET:
            return self.app_cmd_msgHandle
        #else
        return None
        
    def __knowledge(self):
        pass
    
    #util methods
    def getEntities(self) -> list:
        return self.__DE.getEntities()
    
    
    def __isNewSession(self) -> bool:
        if self.__lastChatDth is None:
            return False
        # else:
        return True
    
    def app_cmd_msgHandle(self, response, context):
        #print(response)
        parse = ParseResponse(response)
        
        msgReturnList = MISUNDERSTANDING #default
        #celebrity = first_entity_resolved_value(response['entities'], 'wit$notable_person:notable_person')
        #greetings
        if parse.intentIs_GREET():
            msgReturnList = GREETINGS
        elif parse.intentIs_CREATE_OR_UPDATE(): #TODO: #17 refactoring to change code to DomainEngine
            classList = parse.getEntities_CLASS()
            if len(classList) == 0:
                pass # TODO: #5 to set correct error message (use case no indicate class)
            elif len(classList) >= 2:
                pass # TODO: #6 to set correct error message (use case more than one class)
            else: #all rigth. one class use case
                #seeking for attributes
                attList = parse.getEntities_ATTRIBUTE()
                if len(attList) == 0:
                    pass # TODO: #7 to set correct error message (use case no indicate attributes)
                elif len(attList) % 2 == 1: #it's odd
                    pass # TODO: #8 to set correct error message (use case odd att number)
                else: #all ok! even number!
                    #including the entity
                    domain_entity = self.__DE.addEntity(classList[0].body)
                    for i in range(0, len(attList)-1, 2):
                        self.__DE.addAttribute(domain_entity, attList[i].body, 'str') #TODO: #18 to manage the type 
                    self.__IC.updateModel(showLogs=False) 
                    msgReturnList = CREATE_OR_UPDATE_SUCCESS
        elif parse.intentIs_DELETE(): 
            pass #TODO: #9 elif parse.intentIs_DELETE: 
        elif parse.intentIs_READ(): 
            pass #TODO: #10 elif parse.intentIs_READ:
        elif parse.intentIs_SAY_GOODBYE(): 
            msgReturnList = BYE

        return random.choice(msgReturnList)





from src.text2system.aiengine import AIEngine
from config import SUFIX_CONFIG, SUFIX_ENV, SUFIX_WEB
from src.text2system.analyticsengine import AnalyticsEngine
from src.text2system.businessprocessengine import BusinessProcessEngine
import os
import subprocess as sp
import fileinput
import platform
from config import MANAGED_SYSTEM_NAME

class InterfaceController:
    def __init__(self, AC): #TODO: #4 to analyze the bidirectional relation
        self.__AC = AC #Autonomous Controller Object 
        self.__AIE = AIEngine() #relation 8.1
        self.__BPE = BusinessProcessEngine(self) #relation 8.2
        self.__AE = AnalyticsEngine(self) #relation 8.3
        
        self.__root_path = os.path.dirname(os.path.dirname(__file__)) #get the parent directory
        os.chdir(self.__root_path)

        #starting the python virtual env
        #https://docs.python.org/3/tutorial/venv.html
        self.__venv_path = self.__checkPath(MANAGED_SYSTEM_NAME + SUFIX_ENV)
        
        if not os.path.exists(self.__venv_path):
            print('Creating the python virtual environment...')
            self.__runSyncCmd('python -m venv ' + self.__venv_path) #synchronous
            
        print('Activating the python virtual environment...')
        os.chdir(self.__venv_path) #will stay all runtime in this dir

        if self.__isWindowsServer():
            self.__runSyncCmd('Scripts\\activate.bat')
        else:
            self.__runSyncCmd('. bin/activate')

        #updating o pip
        self.__runSyncCmd('Scripts\\python.exe -m pip install --upgrade pip')
        #install django in virtual environment
        self.__runSyncCmd('Scripts\\pip.exe install django')
        self.__runSyncCmd('Scripts\\pip.exe install django-livesync')
        
        print('creating django config dir...')
        self.__config_path = self.__checkPath(MANAGED_SYSTEM_NAME + SUFIX_CONFIG) 
        self.__settings_path = self.__checkPath(self.__config_path + '\\' + self.__config_path + '\\settings.py')
        if not os.path.exists(self.__config_path):
            print('starting django project...')
            self.__runSyncCmd('Scripts\\django-admin.exe startproject ' + self.__config_path) #synchronous
            print('starting django project (finished)...')

        self.__webapp_path = MANAGED_SYSTEM_NAME + SUFIX_WEB

        if not os.path.exists(self.__webapp_path):
            print('updating manage.py file...')
            self.__runSyncCmd('Scripts\\python.exe  ' + self.__config_path + '\\manage.py startapp ' + self.__webapp_path)  #synchronous
            #extra setup in settings.py
            for line in fileinput.FileInput(self.__settings_path,inplace=1):
                if "    'django.contrib.staticfiles'," in line:
                    line=line.replace(line, "    'livesync',\n" + line + "    '" + self.__webapp_path 
                                      + '.apps.' + self.__webapp_path.replace('_',' ').title().replace(' ','')
                                      + "Config',")
                elif "MIDDLEWARE = [" in line:
                    line = line.replace(line, 
                                        "MIDDLEWARE_CLASSES = ('livesync.core.middleware.DjangoLiveSyncMiddleware')\n\n"
                                        + line)
                elif "ALLOWED_HOSTS = []" in line:
                    line = line.replace(line, "ALLOWED_HOSTS = ['*',]") #for thsi version, allow all hosts                
                print(line, end='')
            print('updating manage.py file...(done)')
            fileinput.close()
            self.migrateModel()
            #creating superuser
            #needs creating the follow system variables:
            #https://stackoverflow.com/questions/26963444/django-create-superuser-from-batch-file/26963549
            '''
            os.environ['DJANGO_SUPERUSER_USERNAME'] = '<<some username>>'
            os.environ['DJANGO_SUPERUSER_PASSWORD'] = '<<some password>>'
            os.environ['DJANGO_SUPERUSER_EMAIL'] = '<<some email>>'
            '''
            print('creating Django superuser...')
            self.__runSyncCmd('Scripts\\python.exe ' + self.__config_path + '\\manage.py createsuperuser --noinput') #--username=root --email=andersonmg@gmail.com')
            
        self.migrateModel()

    def __runServer(self):
        print('running server')
        os.chdir(self.__config_path)
        self.__runAsyncCmd('..\\Scripts\\python.exe manage.py runserver 0.0.0.0:80')# --noreload')       
        os.chdir(self.__checkPath('..\\'))
        return True
    
    def getApp_cmd(self, msgHandle):
        return self.__AIE.getNLPEngine().interactive(msgHandle)        
        
    def updateModel(self, showLogs = True):
        #update admin.py
        if showLogs:
            print('updating admin.py...')
            
        strFileBuffer = 'from django.contrib import admin\nfrom .models import *\n'
        for entity in self.__getEntities():
            #only add entities with one attribute at least
            if len(entity.getAttributes())==0:
                continue
            strFileBuffer += '\n' + f'admin.site.register({entity.name.capitalize()})'

        if not self.__updateFile(self.__webapp_path + '\\admin.py', strFileBuffer):
            return False
        #else:
                
        #update models.py
        if showLogs:
            print('updating models.py...')
        
        strFileBuffer = 'from django.db import models\n'
        for entity in self.__getEntities():
            #only add entities with one attribute at least
            if len(entity.getAttributes())==0:
                continue
            #else: entitie with attributes
            strFileBuffer += '\n' + 'class ' + entity.name.capitalize() + '(models.Model):'
            for att in entity.getAttributes():
                #all fiels with the same type, in this version.
                strFileBuffer += f'\n    {att.name} = models.CharField(max_length=200, null={not att.notnull}, blank={not att.notnull})' 
        #print(strFileBuffer)
        #re-writing the model.py file
        
        if not self.__updateFile(self.__webapp_path + '\\models.py', strFileBuffer):
            return False
        #else:
                
        self.migrateModel(showLogs)
        return True #TODO: #21 to analyse the type of return
        
    
    def updateAppWeb(self):
        return self.updateModel() and self.__runServer()

    
    #util methods
    def __getEntities(self) -> list:
        return self.__AC.getEntities()
        
    def __updateFile(self, path, txtContent):
        if not os.access(path, os.R_OK):
            return False
        
        with open(path, 'w', encoding='utf-8') as f:
            f.write(txtContent)
            f.close()
        return True     
        
    def migrateModel(self, showLogs=True):
        if showLogs:
            print('migrating model...')
        
        self.__runSyncCmd('Scripts\\python.exe ' + self.__config_path + '\\manage.py makemigrations ' + self.__webapp_path)
        self.__runSyncCmd('Scripts\\python.exe ' + self.__config_path + '\\manage.py migrate')
        
    def __isWindowsServer(self) -> bool:
        return platform.system() == 'Windows'
    
    def __checkPath(self, strCmd) -> str:
        checkedStrCmd = os.path.join(*strCmd.split('\\'))
        if not self.__isWindowsServer():
            checkedStrCmd = checkedStrCmd.replace('.exe', '')
            checkedStrCmd = checkedStrCmd.replace('.bat', '')
            checkedStrCmd = checkedStrCmd.replace('Scripts', 'bin')
            
        return checkedStrCmd
    
    def __runSyncCmd(self, strCmd) -> None:
        os.system(self.__checkPath(strCmd))
                
    def __runAsyncCmd(self, strCmd):
        return sp.Popen(self.__checkPath(strCmd).split(), #asynchronous
                                stdout=sp.PIPE,
                                universal_newlines=True, shell=True)        



import os
from src.text2system.multchannelapp import MultChannelApp

class User:
    def __init__(self, login, pwd):
        self.MUP = MultChannelApp(self) 
        self.login = login
        self.__pwd = pwd
    
    #util methods
    @staticmethod
    def getRandomNewUser():
        return User(os.environ['DJANGO_SUPERUSER_USERNAME'], os.environ['DJANGO_SUPERUSER_PASSWORD'])



from src.text2system.auxiliary.constants import *
from src.text2system.securityengine import SecurityEngine
import webbrowser

class MultChannelApp:
    def __init__(self, user):
        self.__user = user #same user for all
        self.__SE = SecurityEngine(self) #security engine instance
    
    def runApp_web(self):
        return webbrowser.open(self.__SE.execute(OPR_APP_HOME_WEB)['homeurl'])
    
    def runApp_cmd(self):
        self.__SE.execute(OPR_APP_HOME_CMD)
    
    def getBotHandle(self):
        return self.__SE.execute(OPR_BOT_HANDLE_GET)        
        
    #CRUD data operations
    def addData(self, data): 
        pass
    def updateData(self, data): 
        pass
    def delData(self, data): 
        pass
    def selectData(self): 
        pass
    
    #Meta-data operations
    def addEntity(self, name):
        return self.__SE.execute(OPR_ENTITY_ADD, {'name' : name})
    
    def addAttribute(self, entity, name, type, notnull=False):
        return self.__SE.execute(OPR_ATTRIBUTE_ADD, {'entity':entity, 'name':name
                                                     , 'type':type, 'notnull':notnull})
       



from src.text2system.auxiliary.attribute import Attribute

class Entity: #TODO: #12 choose other name to differ from responseParse.entity
    def __init__(self, name):
        self.name = name
        self.__attributes = [] #TODO: #20 to change to set()
        #self.addAttribute('att0', 'str', True) #all entities need at least one attribute not null

    def getAttributes(self):
        return self.__attributes
    
    def addAttribute(self, name, type, notNull=False):
        new_att = Attribute(self, name, type, notNull)
        #if self.getAttributes().count(new_att) > 0:
        #    return None
        #else
        self.__attributes.append(new_att)
        
    def delAttribute(self, name):
        self.__attributes.remove(name)
        
    def __eq__(self, o: object) -> bool:
        if type(o) == type(''):
            return self.name == o
        elif type(o) == type(self):
            return self.name == o.name
        #else
        return False         



class Attribute:
    def __init__(self, entity, name, type, notNull=False):
        self.__entity = entity
        self.name = name
        self.type = type
        self.notnull = notNull
        
    def __eq__(self, o: object) -> bool:
        if type(o) == type(''):
            return self.name == o
        elif type(o) == type(self):
            return self.name == o.name
        #else
        return False 



from config import WIT_ACCESS_KEY
from wit import Wit

class AIEngine:
    def __init__(self):
        self.__WIT_CLIENT = None

    #AI Services
    def getNLPEngine(self):
        if self.__WIT_CLIENT == None:
            self.__WIT_CLIENT = Wit(access_token=WIT_ACCESS_KEY)
        return self.__WIT_CLIENT

